# encoding: utf-8

# SNAGED FROM https://github.com/pawelzny/dotty_dict/blob/98984911a61ae9f1aa4da3f6c4808da991b89847/tests/test_dotty_api.py
# UNDER THe MIT LICENSE

import mo_dots
from mo_dots import split_field, to_data, Data, from_data
from mo_testing.fuzzytestcase import FuzzyTestCase


class TestDottyPrivateMembers(FuzzyTestCase):
    def test_split_separator(self):
        result = split_field("chain.of.keys")
        self.assertEqual(result, ["chain", "of", "keys"])


class TestDottyPublicMembers(FuzzyTestCase):
    def test_to_dict(self):
        plain_dict = {"very": {"deeply": {"nested": {"thing": "spam"}}}}
        dot = to_data(plain_dict)
        self.assertIsInstance(from_data(dot), dict)
        self.assertEqual(sorted(from_data(dot).items()), sorted(plain_dict.items()))
        self.assertEqual(sorted(dot.items()), sorted(plain_dict.items()))

    def test_nested_dotty_object_to_dict(self):
        expected_dict = {"hello": {"world": 1}, "nested": {"to_data": {"wazaa": 3}}}
        top_dot = to_data({"hello": {"world": 1}})
        nested_dot = to_data({"wazaa": 3})
        top_dot["nested.to_data"] = nested_dot
        self.assertEqual(from_data(top_dot), expected_dict)

    def test_nested_dotty_in_list_to_dict(self):
        expected_dict = {"testlist": [{"dot1": 1}, {"dot2": 2}]}
        dot_list = [to_data({"dot1": 1}), to_data({"dot2": 2})]
        top_dot = to_data({"testlist": dot_list})
        self.assertEqual(from_data(top_dot), expected_dict)


class TestDictSpecificMethods(FuzzyTestCase):
    def setUp(self):
        self.dot = to_data({
            "flat_key": "flat value",
            "deep": {"nested": 12, "deeper": {"secret": "abcd", "ridiculous": {"hell": "is here", }, }, },
        })

    def test_access_keys(self):
        keys = list(sorted(self.dot.keys()))
        self.assertEqual(keys, ["deep", "flat_key"])

    def test_access_keys_from_deeply_nested_structure(self):
        keys = list(sorted(self.dot["deep.deeper"].keys()))
        self.assertEqual(keys, ["ridiculous", "secret"])

    def test_get_value_without_default(self):
        result = self.dot.get("deep.nested")
        self.assertEqual(result, 12)

    def test_get_value_with_default(self):
        result = self.dot.get("deep.other", False)
        self.assertFalse(result)

    def test_return_dotty_length(self):
        self.assertEqual(len(self.dot), 2)

    def test_pop_from_dotty_flat(self):
        result = self.dot.pop("flat_key")
        self.assertEqual(result, "flat value")
        self.assertEqual(
            self.dot, {"deep": {"nested": 12, "deeper": {"secret": "abcd", "ridiculous": {"hell": "is here", }, }, }, },
        )

    def test_pop_with_default_value(self):
        result = self.dot.pop("not_existing", "abcd")
        self.assertEqual(result, "abcd")
        self.assertEqual(
            self.dot,
            {
                "flat_key": "flat value",
                "deep": {"nested": 12, "deeper": {"secret": "abcd", "ridiculous": {"hell": "is here", }, }, },
            },
        )

    def test_pop_nested_key(self):
        result = self.dot.pop("deep.nested")
        self.assertEqual(result, 12)
        self.assertEqual(
            self.dot,
            {"flat_key": "flat value",
             "deep": {"deeper": {"secret": "abcd", "ridiculous": {"hell": "is here", }, }, }, },
        )

    def test_pop_nested_key_with_default_value(self):
        result = self.dot.pop("deep.deeper.not_existing", "abcd")
        self.assertEqual(result, "abcd")
        self.assertEqual(
            self.dot,
            {
                "flat_key": "flat value",
                "deep": {"nested": 12, "deeper": {"secret": "abcd", "ridiculous": {"hell": "is here", }, }, },
            },
        )

    def test_setdefault_flat_not_existing(self):
        result = self.dot.setdefault("next_flat", "new default value")
        self.assertEqual(result, "new default value")
        self.assertEqual(
            self.dot,
            {
                "flat_key": "flat value",
                "next_flat": "new default value",
                "deep": {"nested": 12, "deeper": {"secret": "abcd", "ridiculous": {"hell": "is here", }, }, },
            },
        )

    def test_setdefault_flat_existing(self):
        self.dot["next_flat"] = "original value"
        result = self.dot.setdefault("next_flat", "new default value")
        self.assertEqual(result, "original value")
        self.assertEqual(
            self.dot,
            {
                "flat_key": "flat value",
                "next_flat": "original value",
                "deep": {"nested": 12, "deeper": {"secret": "abcd", "ridiculous": {"hell": "is here", }, }, },
            },
        )

    def test_setdefault_nested_key_not_existing(self):
        result = self.dot.setdefault("deep.deeper.next_key", "new default value")
        self.assertEqual(result, "new default value")
        self.assertEqual(
            self.dot,
            {
                "flat_key": "flat value",
                "deep": {
                    "nested": 12,
                    "deeper": {"next_key": "new default value", "secret": "abcd",
                               "ridiculous": {"hell": "is here", }, },
                },
            },
        )

    def test_setdefault_nested_key_existing(self):
        self.dot["deep.deeper.next_key"] = "original value"
        result = self.dot.setdefault("deep.deeper.next_key", "new default value")
        self.assertEqual(result, "original value")
        self.assertEqual(
            self.dot,
            {
                "flat_key": "flat value",
                "deep": {
                    "nested": 12,
                    "deeper": {"next_key": "original value", "secret": "abcd", "ridiculous": {"hell": "is here", }, },
                },
            },
        )

    def test_copy(self):
        first = to_data({"a": 1, "b": 2})
        second = first.copy()

        self.assertIsInstance(second, Data)
        self.assertEqual(first, second)
        self.assertIsNot(first, second)
        self.assertIsNot(first._data, second._data)

    def test_fromkeys1(self):
        dot = mo_dots.fromkeys({"a", "b", "c"}, value=10)
        self.assertEqual(from_data(dot), {"a": 10, "b": 10, "c": 10})
        self.assertIsInstance(dot, Data)

    def test_fromkeys2(self):
        dot = mo_dots.fromkeys({"a", "b", "c"}, value=None)
        self.assertEqual(len(from_data(dot).keys()), 0)



# encoding: utf-8
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this file,
# You can obtain one at http://mozilla.org/MPL/2.0/.
#
# Contact: Kyle Lahnakoski (kyle@lahnakoski.com)
#
from copy import deepcopy, copy
from math import isnan

from mo_future import UserDict, first, Mapping
from mo_logs import Log
from mo_logs.strings import expand_template
from mo_math import MAX
from mo_testing.fuzzytestcase import FuzzyTestCase

from mo_dots import (
    to_data,
    Null,
    set_default,
    Data,
    literal_field,
    NullType,
    leaves_to_data,
    from_data,
    FlatList,
    get_attr,
    relative_field,
    unliteral_field,
    tail_field,
    join_field,
    set_attr,
    PATH_NOT_FOUND,
)
from mo_dots.datas import leaves
from mo_dots.objects import datawrap
from tests import ambiguous_test


class TestDot(FuzzyTestCase):
    def test_set_union_w_null(self):
        s = set("a")
        s |= Null
        self.assertAlmostEqual(s, set("a"))

    def test_null_class(self):
        self.assertFalse(isinstance(Null, Mapping))

    def test_userdict_1(self):
        def show_kwargs(**kwargs):
            return kwargs

        a = UserDict(a=1, b=2)
        d = show_kwargs(**a)
        self.assertAlmostEqual(d, {"a": 1, "b": 2})

    def test_userdict_2(self):
        def show_kwargs(**kwargs):
            return kwargs

        a = UserDict()
        a.data["a"] = 1
        a.data["b"] = 2
        d = show_kwargs(**a)
        self.assertAlmostEqual(d, {"a": 1, "b": 2})

    def test_dict_args(self):
        def show_kwargs(**kwargs):
            return kwargs

        a = Data()
        a["a"] = 1
        a["b"] = 2
        d = show_kwargs(**a)
        self.assertAlmostEqual(d, {"a": 1, "b": 2})

    def test_is_mapping(self):
        self.assertTrue(isinstance(Data(), Mapping), "All Data must be Mappings")

    def test_kwargs(self):
        d = Data(a=1, b=2)

        def func(a, b):
            return a == 1, b == 2

        self.assertTrue(func(**d))

    def test_none(self):
        a = 0
        b = 0
        c = None
        d = None

        if a == b:
            pass
        else:
            Log.error("error")

        if c == d:
            pass
        else:
            Log.error("error")

        if a == c:
            Log.error("error")

        if d == b:
            Log.error("error")

        if not c:
            pass
        else:
            Log.error("error")

    def test_null_access(self):
        a = Data()
        c = a.b["test"]
        self.assertTrue(c == None)

    def test_null(self):
        a = 0
        b = 0
        c = Null
        d = Null
        e = Data()
        f = Data()

        if a == b:
            pass
        else:
            Log.error("error")

        if c == d:
            pass
        else:
            Log.error("error")

        if a == c:
            Log.error("error")

        if d == b:
            Log.error("error")

        if c == None:
            pass
        else:
            Log.error("error")

        if not c:
            pass
        else:
            Log.error("error")

        if Null != Null:
            Log.error("error")

        if Null != None:
            Log.error("error")

        if None != Null:
            Log.error("error")

        if e.test != f.test:
            Log.error("error")

    def test_get_value(self):
        a = to_data({"a": 1, "b": {}})

        if a.a != 1:
            Log.error("error")
        if not isinstance(a.b, Mapping):
            Log.error("error")

    def test_get_class(self):
        a = to_data({})
        _type = a.__class__

        if _type is not Data:
            Log.error("error")

    def test_int_null(self):
        a = Data()
        value = a.b * 1000
        assert value == Null

    def test_dot_self(self):
        a = Data(b=42)
        assert a["."] == a
        assert a["."].b == 42

        a["."] = {"c": 42}
        assert a.c == 42
        assert a.b == None

    def test_list(self):
        if not []:
            pass
        else:
            Log.error("error")

        if []:
            Log.error("error")

        if not [0]:
            Log.error("error")

    def test_assign1(self):
        a = {}

        b = to_data(a)
        b.c = "test1"
        b.d.e = "test2"
        b.f.g.h = "test3"
        b.f.i = "test4"
        b.k["l.m.n"] = "test5"

        expected = {
            "c": "test1",
            "d": {"e": "test2"},
            "f": {"g": {"h": "test3"}, "i": "test4"},
            "k": {"l": {"m": {"n": "test5"}}},
        }
        self.assertEqual(a, expected)

    def test_assign2(self):
        a = {}

        b = to_data(a)
        b_c = b.c
        b.c.d = "test1"

        b_c.e = "test2"

        expected = {"c": {"d": "test1", "e": "test2"}}
        self.assertEqual(a, expected)

    def test_assign3(self):
        # IMPOTENT ASSIGNMENTS DO NOTHING
        a = {}
        b = to_data(a)

        b.c = None
        expected = {}
        self.assertEqual(a, expected)

        b.c.d = None
        expected = {}
        self.assertEqual(a, expected)

        b["c.d"] = None
        expected = {}
        self.assertEqual(a, expected)

        b.c.d.e = None
        expected = {}
        self.assertEqual(a, expected)

        b.c["d.e"] = None
        expected = {}
        self.assertEqual(a, expected)

    def test_assign4(self):
        # IMPOTENT ASSIGNMENTS DO NOTHING
        a = {"c": {"d": {}}}
        b = to_data(a)
        b.c.d = None
        expected = {"c": {}}
        self.assertEqual(a, expected)

        a = {"c": {"d": {}}}
        b = to_data(a)
        b.c = None
        expected = {}
        self.assertEqual(a, expected)

    def test_assign5(self):
        a = {}
        b = to_data(a)

        b.c["d..e"].f = 2
        expected = {"c": {"d.e": {"f": 2}}}
        self.assertEqual(a, expected)

    def test_assign6(self):
        a = {}
        b = to_data(a)

        b["c.d.e..f"] = 1
        b["c.d.e..g"] = 2

        expected = {"c": {"d": {"e.f": 1, "e.g": 2}}}
        self.assertEqual(a, expected)

    def test_assign7(self):
        a = {}
        b = to_data(a)

        b["c.d.e..f"] = 1
        b["c.d.g..h"] = 2

        expected = {"c": {"d": {"e.f": 1, "g.h": 2}}}
        self.assertEqual(a, expected)

    def test_assign8(self):
        a = {}
        b = to_data(a)

        b["a"][literal_field(literal_field("b.html"))]["z"] = 3

        expected = {"a": {"b..html": {"z": 3}}}
        self.assertEqual(a, expected)

    def test_assign9(self):
        a = {}
        b = to_data(a)

        b["a"]["."] = 1

        expected = {"a": 1}
        self.assertEqual(a, expected)

    def test_setitem_and_deep(self):
        a = {}
        b = to_data(a)

        b.c["d"].e.f = 3
        expected = {"c": {"d": {"e": {"f": 3}}}}
        self.assertEqual(a, expected)

    def test_assign_and_use1(self):
        a = to_data({})
        agg = a.b
        agg.c = []

        v = agg.c
        agg.c.append("test value")

        self.assertEqual(a, {"b": {"c": ["test value"]}})
        self.assertEqual(a.b, {"c": ["test value"]})
        self.assertEqual(a.b.c, ["test value"])

    def test_assign_and_use2(self):
        a = to_data({})
        agg = a.b.c
        agg += []
        agg.append("test value")

        self.assertEqual(a, {"b": {"c": ["test value"]}})
        self.assertEqual(a.b, {"c": ["test value"]})
        self.assertEqual(a.b.c, ["test value"])

    def test_assign_and_use3(self):
        a = to_data({})
        agg = a.b
        agg.c += ["test value"]

        self.assertEqual(a, {"b": {"c": ["test value"]}})
        self.assertEqual(a.b, {"c": ["test value"]})
        self.assertEqual(a.b.c, ["test value"])

    def test_assign_none(self):
        a = {}
        A = to_data(a)

        A[None] = "test"
        self.assertEqual(a, {})

    def test_increment(self):
        a = {}
        b = to_data(a)
        b.c1.d += 1
        b.c2.e += "e"
        b.c3.f += ["f"]
        b["c..a"].d += 1

        self.assertEqual(a, {"c1": {"d": 1}, "c2": {"e": "e"}, "c3": {"f": ["f"]}, "c.a": {"d": 1}})

        b.c1.d += 2
        b.c2.e += "f"
        b.c3.f += ["g"]
        b["c..a"].d += 3
        self.assertEqual(
            a, {"c1": {"d": 3}, "c2": {"e": "ef"}, "c3": {"f": ["f", "g"]}, "c.a": {"d": 4},},
        )

    def test_slicing(self):
        def diff(record, index, records):
            """
            WINDOW FUNCTIONS TAKE THE CURRENT record, THE index THAT RECORD HAS
            IN THE WINDOW, AND THE (SORTED) LIST OF ALL records
            """
            # COMPARE CURRENT VALUE TO MAX OF PAST 5, BUT NOT THE VERY LAST ONE
            try:
                return record - MAX(records[index - 6 : index - 1 :])
            except Exception as e:
                return None

        data1_list = [1, 2, 3, 4, 5, 6, 7, 8, 9]
        result1 = [diff(r, i, data1_list) for i, r in enumerate(data1_list)]
        assert result1 == [
            -7,
            None,
            None,
            None,
            None,
            None,
            2,
            2,
            2,
        ]  # WHAT IS EXPECTED, BUT NOT WHAT WE WANT

        data2_list = to_data(data1_list)
        result2 = [diff(r, i, data2_list) for i, r in enumerate(data2_list)]
        assert result2 == [None, None, 2, 2, 2, 2, 2, 2, 2]

    def test_delete1(self):
        a = to_data({"b": {"c": 1}})

        del a.b.c
        self.assertEqual({"b": {}}, a)
        self.assertEqual(a, {"b": {}})

        a = to_data({"b": {"c": 1}})

        a.b.c = None
        self.assertEqual({"b": {}}, a)
        self.assertEqual(a, {"b": {}})

    def test_delete2(self):
        a = to_data({"b": {"c": 1, "d": 2}})

        del a.b.c
        self.assertEqual({"b": {"d": 2}}, a)
        self.assertEqual(a, {"b": {"d": 2}})
        a = to_data({"b": {"c": 1, "d": 2}})

        a.b.c = None
        self.assertEqual({"b": {"d": 2}}, a)
        self.assertEqual(a, {"b": {"d": 2}})

    def test_wrap(self):
        d = {}
        dd = to_data(d)
        self.assertIs(from_data(dd), d)

    def test_object_wrap(self):
        d = SampleData()
        dd = datawrap(d)

        self.assertEqual(dd["a"], 20)
        self.assertEqual(dd, {"a": 20, "b": 30})
        self.assertIs(from_data(dd), d)

    def test_object_wrap_w_deep_path(self):
        d = SampleData()
        d.a = Data(c=3)
        dd = datawrap(d)

        self.assertEqual(dd["a.c"], 3)
        self.assertEqual(dd, {"a": {"c": 3}, "b": 30})

    def test_deep_select(self):
        d = to_data([{"a": {"b": 1}}, {"a": {"b": 2}}])

        test = d.a.b
        self.assertEqual(test, [1, 2])

    def test_deep_select_list(self):
        d = to_data({"a": {"b": [{"c": 1}, {"c": 2}]}})

        test = d["a.b.c"]
        self.assertEqual(test, [1, 2])

    def test_set_default1(self):
        a = {"x": {"y": 1}}
        b = {"x": {"z": 2, "y": None}}
        c = {}
        d = set_default(c, a, b)

        self.assertIs(from_data(d), c)
        self.assertEqual(d.x.y, 1, "expecting d to have attributes of a")
        self.assertEqual(d.x.z, 2, "expecting d to have attributes of b")

        self.assertEqual(to_data(a).x.z, None, "a should not have been altered")

    def test_set_default2(self):
        a = {"x": {"y": 1}}
        b = {"x": {"z": 2}}
        c = {}
        d = set_default(c, None, a, b)

        self.assertIs(from_data(d), c)
        self.assertEqual(d.x.y, 1, "expecting d to have attributes of a")
        self.assertEqual(d.x.z, 2, "expecting d to have attributes of b")

        self.assertEqual(to_data(a).x.z, None, "a should not have been altered")

    def test_set_default3(self):
        a = {"x": {"y": 1}}
        b = {"x": {"z": 2}}
        c = {}
        d = set_default(None, c, a, b)

        self.assertIsNot(from_data(d), c)
        self.assertEqual(d.x.y, 1, "expecting d to have attributes of a")
        self.assertEqual(d.x.z, 2, "expecting d to have attributes of b")

        self.assertEqual(to_data(a).x.z, None, "a should not have been altered")

    def test_or(self):
        a = {"x": {"y": 1}}
        b = {"x": {"z": 2}}
        c = Data()
        d = c | a | b

        self.assertEqual(d.x.y, 1, "expecting d to have attributes of a")
        self.assertEqual(d.x.z, 2, "expecting d to have attributes of b")

        self.assertEqual(to_data(a).x.z, None, "a should not have been altered")

    def test_Data_of_Data(self):
        value = {"a": 1}
        wrapped = to_data(to_data(value))
        self.assertTrue(value is from_data(wrapped), "expecting identical object")

    def test_leaves_of_mappings(self):
        a = to_data({"a": _TestMapping()})
        a.a.a = {"a": 1}
        a.a.b = {"b": 2}

        leaves = to_data(dict(a.leaves()))
        self.assertEqual(a.a.a["a"], leaves["a..a..a"], "expecting 1")
        self.assertEqual(a.a.b["b"], leaves["a..b..b"], "expecting 2")

    def test_null_set_index(self):
        temp = Null
        # expecting no crash
        temp[0] = 1
        temp[1] = None

    def test_deep_null_assignment(self):
        temp = to_data({"a": 0})
        e = temp.e
        e.s.t = 1
        e.s.s = 2
        self.assertEqual(temp, {"a": 0, "e": {"s": {"s": 2, "t": 1}}}, "expecting identical")

    def test_null_inequalities(self):
        self.assertEqual(Null < 1, None)
        self.assertEqual(Null <= 1, None)
        self.assertEqual(Null != 1, None)
        self.assertEqual(Null == 1, None)
        self.assertEqual(Null >= 1, None)
        self.assertEqual(Null > 1, None)

        self.assertEqual(1 < Null, None)
        self.assertEqual(1 <= Null, None)
        self.assertEqual(1 != Null, None)
        self.assertEqual(1 == Null, None)
        self.assertEqual(1 >= Null, None)
        self.assertEqual(1 > Null, None)

        self.assertEqual(Null < Null, None)
        self.assertEqual(Null <= Null, None)
        self.assertEqual(Null != Null, False)
        self.assertEqual(Null == Null, True)
        self.assertEqual(Null >= Null, None)
        self.assertEqual(Null > Null, None)

        self.assertEqual(Null < None, None)
        self.assertEqual(Null <= None, None)
        self.assertEqual(Null != None, False)
        self.assertEqual(Null == None, True)
        self.assertEqual(Null >= None, None)
        self.assertEqual(Null > None, None)

        self.assertEqual(None < Null, None)
        self.assertEqual(None <= Null, None)
        self.assertEqual(None != Null, False)
        self.assertEqual(None == Null, True)
        self.assertEqual(None >= Null, None)
        self.assertEqual(None > Null, None)

    def test_escape_dot(self):
        self.assertAlmostEqual(literal_field("."), "\b")
        self.assertAlmostEqual(literal_field(".."), "\b\b")
        self.assertAlmostEqual(literal_field("..."), "\b..\b")
        self.assertAlmostEqual(literal_field("a.b"), "a..b")
        self.assertAlmostEqual(literal_field("a..html"), "a....html")

    def test_set_default_unicode_and_list(self):
        a = {"a": "test"}
        b = {"a": [1, 2]}
        self.assertAlmostEqual(
            set_default(a, b), {"a": ["test", 1, 2]}, "expecting string, not list, nor some hybrid",
        )

    def test_unicode_or_list(self):
        a = to_data({"a": "test"})
        b = {"a": [1, 2]}
        self.assertAlmostEqual(a | b, {"a": "test"}, "expecting string, not list, nor some hybrid")
        self.assertAlmostEqual(b | a, {"a": [1, 2]}, "expecting list")

    def test_deepcopy(self):
        self.assertIs(deepcopy(Null), Null)
        self.assertEqual(deepcopy(Data()), {})
        self.assertEqual(deepcopy(Data(a=Null)), {})

    def test_null_type(self):
        self.assertIs(Null.__class__, NullType)
        self.assertTrue(isinstance(Null, NullType))

    def test_null_assign(self):
        output = Null
        output.changeset.files = None

    def test_string_assign(self):
        def test():
            a = to_data({"a": "world"})
            a["a.html"] = "value"

        self.assertRaises(Exception, test, "expecting error")

    def test_string_assign_null(self):
        a = to_data({"a": "world"})
        a["a.html"] = None

    def test_empty_object_is_not_null1(self):
        self.assertFalse(to_data({}) == None, "expect empty objects to not equal None")

    def test_empty_object_is_not_null2(self):
        self.assertTrue(to_data({}) != None, "expect empty objects to not equal None")

    def test_add_null_to_list(self):
        expected = to_data(["test", "list"])
        test = expected + None
        self.assertEqual(test, expected, "expecting adding None to list does not change list")

    def test_pop_list(self):
        l = to_data([1, 2, 3, 4])

        self.assertEqual(l.pop(3), 4)
        self.assertEqual(l.pop(0), 1)
        self.assertEqual(l.pop(1), 3)
        self.assertEqual(l.pop(), 2)

    def test_pop_dict(self):
        d = to_data({"a": 1, "b": 2, "c": 3})

        self.assertAlmostEqual(d.pop("a"), 1)
        self.assertAlmostEqual({"b": 2, "c": 3}, d)
        self.assertAlmostEqual(d.pop("c"), 3)
        self.assertAlmostEqual({"b": 2}, d)
        self.assertAlmostEqual(d.pop("b"), 2)
        self.assertAlmostEqual({}, d)

    def test_values(self):
        a = to_data({"a": 1, "b": 2})
        result = []
        for v in a.values():
            result.append(v)

        expected = {1, 2}
        self.assertAlmostEqual(result, expected)

    def test_wrap_wrap(self):
        a = to_data({"a": 1, "b": 2})
        b = to_data(a)

        self.assertIs(a, b, "expecting same object")

    def test_key_in_data(self):
        a = to_data({"key": {}})
        self.assertIn("key", a)

    def test_list_eq(self):
        a = to_data([{"a": 1}])
        b = to_data([])
        c = to_data([])

        self.assertFalse(a == b)
        self.assertTrue(a == a)
        self.assertTrue(b == b)
        self.assertTrue(b == c)
        self.assertTrue(b == None)
        self.assertTrue(b == Null)
        self.assertTrue(None == b)
        self.assertTrue(Null == b)

    def test_add_1(self):
        a = to_data({"a": 1, "y": {"b": 4, "c": [5]}})
        b = to_data({"a": 1, "x": 5, "y": {"b": 6, "c": [12]}})

        expected_ab = {"a": 2, "x": 5, "y": {"b": 10, "c": [5, 12]}}
        expected_ba = {"a": 2, "x": 5, "y": {"b": 10, "c": [12, 5]}}
        self.assertEqual(a + b, expected_ab)
        self.assertEqual(b + a, expected_ba)

        a += b
        self.assertEqual(a, expected_ab)

    def test_add_2(self):
        a = to_data({"a": 1, "y": {"b": 4, "c": [5]}})
        b = to_data({"a": 1, "x": 5, "y": {"b": 6, "c": 12}})

        expected_ab = {"a": 2, "x": 5, "y": {"b": 10, "c": [5, 12]}}
        expected_ba = {"a": 2, "x": 5, "y": {"b": 10, "c": [12, 5]}}
        self.assertEqual(a + b, expected_ab)
        self.assertEqual(b + a, expected_ba)

        a += b
        self.assertEqual(a, expected_ab)

    def test_list_get(self):
        flat_list = to_data([{"a": 1}, {"a": None}])
        # THIS IS NOT AN OPTION BECAUSE [] IS RESERVED FOR INDEXING AND SLICING
        self.assertEqual(flat_list["a"], None)

    def test_copy1(self):
        a = Data(b="c")
        aa = a.copy()
        self.assertEqual(aa, a, "expecting to be the same")
        self.assertEqual(a, aa, "expecting to be the same")

        a.b = "d"
        self.assertNotEqual(a, aa, "expecting to be different now")
        self.assertNotEqual(aa, a, "expecting to be different now")

    def test_copy2(self):
        a = Data(b="c")
        aa = copy(a)
        self.assertEqual(aa, a, "expecting to be the same")
        self.assertEqual(a, aa, "expecting to be the same")

        a.b = "d"
        self.assertNotEqual(a, aa, "expecting to be different now")
        self.assertNotEqual(aa, a, "expecting to be different now")

    def test_copy_value(self):
        a = to_data({})
        a["."] = "test"

        aa = a.copy()
        self.assertEqual(aa, a, "expecting to be the same")
        self.assertEqual(a, aa, "expecting to be the same")

    def test_in(self):
        a = {"_id": "yes"}
        b = {"id": "no"}
        aa = to_data(a)
        bb = to_data(b)

        self.assertEqual("_id" in aa, "_id" in a)
        self.assertEqual("_id" in bb, "_id" in b)

    def test_in_null(self):
        self.assertIn(Null, [None])

    def test_in_none(self):
        self.assertIn(None, [Null])

    def test_none_and_null_in_dict(self):
        d = {None: None, Null: None}
        self.assertEqual(len(d), 1)

    def test_none_and_null_compare_with_list(self):
        empty = to_data([])

        self.assertTrue([] == Null)
        self.assertTrue(empty == Null)
        self.assertTrue(empty == None)

    def test_none_and_magic(self):
        # self.assertIs(int(Null), Null)
        # self.assertIs(float(Null), Null)
        self.assertEqual(list(Null), Null)

    def test_keys(self):
        a = to_data({"a.b": "c"})
        self.assertEqual(a.keys(), {"a.b"})

    def test_items(self):
        a = to_data({"a.b": "c"})
        self.assertEqual(a.items(), [("a.b", "c")])

    def test_iteritems(self):
        a = to_data({"a.b": "c"})
        self.assertEqual(list(a.iteritems()), [("a.b", "c")])

    def test_update_complex(self):
        a = Data(a=1, b={"c": 0})
        b = Data(b={"d": 3})
        result = a | b
        self.assertEqual(result, {"a": 1, "b": {"c": 0, "d": 3}})

    def test_string_using_leaves(self):
        result = leaves_to_data("test")
        self.assertNotIsInstance(result, Data)
        self.assertEqual(result, "test")

    def test_leaves_returns_flat_list(self):
        x = to_data({"a": [1, 2, 3]})
        self.assertIsInstance(first(x.leaves())[1], FlatList)

    def test_leaves_returns_flat_list(self):
        x = to_data([{"a": [1, 2, 3]}])
        self.assertIsInstance(x, FlatList)

    def test_leaves_returns_inner(self):
        x = leaves_to_data({"a.b.c": 3, "\b": 42, "d": None})
        self.assertEqual(x, {"a": {"b": {"c": 3}}, ".": 42})

    def test_empty_string_is_bad(self):
        with self.assertRaises(Exception):
            leaves_to_data({"": 4})

    def test_leaves_w_Data(self):
        x = leaves_to_data({"a": to_data({"b.c": 42})})
        self.assertEqual(x, {"a": {"b": {"c": 42}}})

    def test_leaves_w_simple(self):
        x = leaves_to_data(42)
        self.assertEqual(x, 42)

        x = leaves_to_data([42, 0])
        self.assertEqual(x, [42, 0])

    def test_leaves_on_dict(self):
        x = list(leaves({"a": to_data({"b.c": 42})}))
        self.assertEqual(x, [("a.b..c", 42)])

    def test_leaves_on_dict_w_prefix(self):
        x = leaves({"a": to_data({"b.c": 42})}, prefix="::")
        self.assertEqual(x, [("::a.b..c", 42)])

    def test_to_generator(self):
        def gen():
            yield 1
            yield 2

        x = to_data(gen())
        self.assertIsInstance(x, FlatList)
        self.assertEqual(len(x), 2)

    def test_get_module_attr(self):
        x = get_attr(ambiguous_test, "d")
        self.assertEqual(x, {"a": 44})

    def test_get_module_attr_ambiguous(self):
        with self.assertRaises(Exception):
            x = get_attr(ambiguous_test, "DE")

    def test_get_module_attr_lowercase(self):
        x = get_attr(ambiguous_test, "d")
        y = get_attr(ambiguous_test, "D")
        self.assertIs(x, y)

    def test_set_module_attr(self):
        old, Log.main_log = Log.main_log, StructuredLogger_usingList()
        set_attr(ambiguous_test, "d1", "test")
        new, Log.main_log = Log.main_log, old
        self.assertIn(PATH_NOT_FOUND, new.lines[0])
        self.assertEqual(ambiguous_test.d1, "test")

    def test_relative(self):
        self.assertEqual(relative_field("a.b.c", "."), "a.b.c")
        self.assertEqual(relative_field("a.b.c", "a"), "b.c")
        self.assertEqual(relative_field("a.b.c", "a.b"), "c")
        self.assertEqual(relative_field("a.b.c", "a.b.c"), ".")
        self.assertEqual(relative_field("a.b.c", "a.b.c.d"), "..")
        self.assertEqual(relative_field("a.b.c", "a.b.c.d.e"), "...")

        self.assertEqual(relative_field("a.b.c.k", "a.b.c"), "k")
        self.assertEqual(relative_field("a.b.c.k", "a.b.c.d"), "..k")
        self.assertEqual(relative_field("a.b.c.k", "a.b.c.d.e"), "...k")

    def test_unliteral(self):
        for f in (".", "..", "a", "a.b", "..a.b", "", "a.."):
            self.assertEqual(unliteral_field(literal_field(f)), f)

    def test_tail_field(self):
        self.assertEqual(tail_field(None), (".", "."))
        self.assertEqual(tail_field(""), (".", "."))
        self.assertEqual(tail_field("."), (".", "."))
        self.assertEqual(tail_field(".."), ("..", "."))
        self.assertEqual(tail_field("..."), ("..", ".."))
        self.assertEqual(tail_field("...."), ("..", "..."))
        self.assertEqual(tail_field("a"), ("a", "."))
        self.assertEqual(tail_field("a.b"), ("a", "b"))
        self.assertEqual(tail_field("a.b.c."), ("a", "b.c"))

    def test_join_field_generator(self):
        def gen():
            yield "a"
            yield "b"

        self.assertEqual(join_field(gen()), "a.b")

    def test_hash(self):
        k1 = to_data({"a": 1, "b": 2})
        k2 = to_data({"a": 1, "b": 2})
        x = {k1: "42"}

        self.assertEqual(x[k1], x[k2])

        k1 = to_data([{"a": 1}, {"b": 2}])
        k2 = to_data([{"a": 1}, {"b": 2}])
        x = {k1: "42"}

        self.assertEqual(x[k1], x[k2])

    def test_bool(self):
        a = to_data({})
        b = to_data([])
        c = to_data("")
        d = Data()
        d["."] = None
        e = Data()
        e["."] = 0

        self.assertEqual(bool(a), True)
        self.assertEqual(bool(b), False)
        self.assertEqual(bool(c), False)
        self.assertEqual(bool(d), False)
        self.assertEqual(bool(e), True)

    def test_iter(self):
        a = to_data({})
        self.assertEqual(list(a), [])

        b = to_data({"a": 42})
        self.assertEqual(list(iter(b)), [("a", 42)])

    def test_null(self):
        with self.assertRaises(Exception):
            int(Null)

        result = float(Null)
        self.assertTrue(isnan(result))

        self.assertFalse(bool(Null))
        self.assertFalse(len(Null))

        self.assertEqual([1] + Null, [1])
        self.assertEqual(Null + [1], [1])

        self.assertTrue(1 + Null == None)
        self.assertTrue(Null + 1 == None)
        self.assertTrue(Null + Null == None)
        self.assertTrue(Null() == None)
        self.assertTrue(Null - 1 == None)
        self.assertTrue(1 - Null == None)
        self.assertTrue(Null * 1 == None)
        self.assertTrue(1 * Null == None)
        self.assertTrue(Null / 1 == None)
        self.assertTrue(1 / Null == None)
        self.assertTrue(Null / 1.2 == None)
        self.assertTrue(1.2 / Null == None)
        self.assertTrue(Null // 1 == None)
        self.assertTrue(1 // Null == None)
        self.assertTrue(Null | 1 == 1)
        self.assertTrue(1 | Null == 1)
        self.assertTrue(Null ^ 1 == None)
        self.assertTrue(1 ^ Null == None)
        self.assertTrue(Null & 1 == None)
        self.assertTrue(1 & Null == None)
        self.assertFalse(Null & False)
        self.assertFalse(False & Null)
        self.assertTrue((1 > Null) == None)
        self.assertTrue((1 >= Null) == None)
        self.assertTrue((1 <= Null) == None)
        self.assertTrue((1 < Null) == None)
        self.assertTrue(-Null == None)
        self.assertTrue(copy(Null) == Null)
        self.assertTrue(Null[1:3] == Null)
        self.assertTrue(Null[1] == Null)
        self.assertTrue(Null.__div__(0) == None)
        self.assertTrue(Null.__rdiv__(0) == None)
        self.assertTrue(Null.__itruediv__(0) == None)

        x = Null
        x += Null
        self.assertTrue(x == None)

        x = to_data({}).a.b.c
        x += 3
        self.assertTrue(x == 3)

    def test_get_zero(self):
        self.assertEqual(get_attr({"a": [{"b": 2}, 2]}, "a.0.b"), 2)

    def test_set_attr1(self):
        d = {"a": [{"b": 3}, 2]}
        set_attr(d, "a.0.b", 2)
        self.assertEqual(to_data(d).a[0].b, 2)

    def test_set_attr2(self):
        old, Log.main_log = Log.main_log, StructuredLogger_usingList()
        set_attr(None, "a", 2)
        new, Log.main_log = Log.main_log, old
        self.assertIn(PATH_NOT_FOUND, new.lines[0])

    def test_set_attr3(self):
        x = SampleData()
        x.a = SampleData()
        set_attr(x, "a.a", "q")
        self.assertEqual(x.a.a, "q")

        set_attr(x, "a.a", None)
        self.assertEqual(x.a.a, None)

        set_attr(x, "a", 42)
        self.assertEqual(x.a.a, 42)

        x.b = {}
        set_attr(x, "b", None)
        self.assertTrue(x.b == None)

    def test_none_item(self):
        self.assertIsInstance(to_data({"a": 1})[None], NullType)

    def test_assign_to_dot(self):
        x = to_data({"a": 1})
        x["."] = 42
        self.assertEqual(x["."], 42)

    def test_assign_list_to_dot(self):
        x = to_data({"a": 1})
        x["."] = ["a", "b"]
        self.assertEqual(x["."], ["a", "b"])

    def test_repr(self):
        self.assertEqual(repr(to_data({"a": 1})), "to_data({'a': 1})")

    def test_repr2(self):
        d = {}
        for i in range(3):
            d = {"a": d}
        d = to_data(d)
        self.assertEqual(eval(repr(d)), d)

    def test_repr3(self):
        d = Data()
        for i in range(3):
            d = Data(a=d)
        self.assertEqual(repr(d), "to_data({'a': to_data({'a': to_data({'a': to_data({})})})})")

    def test_add_one_to_dict(self):
        x = to_data({"a": {"b": 42}})
        with self.assertRaises(Exception):
            x += {"a": 1}

    def test_add_dict_to_one(self):
        x = to_data({"a": 1})
        with self.assertRaises(Exception):
            x += {"a": {"b": 42}}

    def test_add_dict_to_list(self):
        x = to_data({"a": [1]})
        x += {"a": {"b": 42}}
        self.assertEqual(x, {"a": [1, {"b": 42}]})

    def test_add_str_to_dict(self):
        x = to_data({"a": {"b": 42}})
        with self.assertRaises(Exception):
            x += {"a": "test"}

    def test_add_dict_to_str(self):
        x = to_data({"a": "test"})
        with self.assertRaises(Exception):
            x += {"a": {"b": 42}}

    def test_add_str_to_str(self):
        x = to_data({"a": "test"})
        with self.assertRaises("has no attribute 'append'"):
            x += {"a": "world"}

    def test_iadd(self):
        x = Data()
        x["."]["."] += [42]
        self.assertEqual(x, [42])

    def test_leaves1(self):
        result = leaves_to_data({"a": "hi", "a.b": "test"})
        self.assertEqual(result, {"a": {"b": "test"}})

    def test_leaves2(self):
        result = leaves_to_data({".": "hi"})
        self.assertEqual(result, "hi")

    def test_leaves3(self):
        result = leaves_to_data({".": "hi", "a": 42})
        self.assertEqual(result, {"a": 42})

    def test_leaves4(self):
        result = leaves_to_data({"a": 42, ".": "hi"})
        self.assertEqual(result, {"a": 42})

    def test_leaves5(self):
        data = Data()
        data["."] = "hi"
        b = Data(a=data)

        result = list(b.leaves())
        self.assertEqual(result, [("a", "hi")])

    def test_float(self):
        data = Data()
        data.a = float(data.a)
        self.assertTrue(data.a == None)


class _TestMapping(object):
    def __init__(self):
        self.a = None
        self.b = None


class SampleData(object):
    def __init__(self, a=None):
        self.a = a or 20
        self.b = 30

    def __str__(self):
        return str(self.a) + str(self.b)


class StructuredLogger_usingList(object):
    def __init__(self):
        self.lines = []

    def write(self, template, params):
        self.lines.append(expand_template(template, params))

    def stop(self):
        pass



# encoding: utf-8
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this file,
# You can obtain one at http://mozilla.org/MPL/2.0/.
#
# Contact: Kyle Lahnakoski (kyle@lahnakoski.com)
#

import os
import sys
from collections import deque
from unittest import skipIf

from mo_future import text, Mapping
from mo_logs import Log
from mo_math import randoms
from mo_testing.fuzzytestcase import FuzzyTestCase
from mo_times import Timer

from mo_dots import (
    Data,
    Null,
    FlatList,
    data_types,
    is_data,
    to_data,
    from_data,
    split_field,
    is_null,
)

IS_TRAVIS = bool(os.environ.get("TRAVIS"))
IS_COVERAGE = bool(os.environ.get("COVERAGE"))

PY37 = sys.version_info[0] == 3 and sys.version_info[1] == 7
PY39 = sys.version_info[0] == 3 and sys.version_info[1] == 9


@skipIf(IS_TRAVIS or IS_COVERAGE, "no need to test speed on Travis")
class TestDotSpeed(FuzzyTestCase):
    def test_simple_access(self):
        times = range(1000 * 1000)
        x = to_data({"a": {"b": 42}})
        y = Dummy({"b": 42})

        with Timer("slot access") as slot:
            for i in times:
                k = y.a

        with Timer("attribute access") as att:
            for i in times:
                k = x.a

        with Timer("property access access") as prop:
            for i in times:
                k = x["a"]

        Log.info(
            "attribute access is {{t|round(places=2)}}x faster",
            t=(att.duration.seconds + prop.duration.seconds) / 2 / slot.duration.seconds,
        )

    def test_compare_isinstance_to_class_checks(self):
        num = 1 * 1000 * 1000
        options = {
            0: lambda: {},
            1: lambda: Data(),
            2: lambda: Null,
            3: lambda: 6,
            4: lambda: "string",
        }
        data = [options[randoms.int(len(options))]() for _ in range(num)]

        with Timer("Data: isinstance of Mapping check") as i_time:
            i_result = [isinstance(d, Mapping) for d in data]

        with Timer("Data: in data_types check") as s_time:
            s_result = [d.__class__ in data_types for d in data]

        with Timer("Data: is checks") as e_time:
            e_result = [d.__class__ is Data or d.__class__ is dict for d in data]

        with Timer("Data: is_instance checks") as n_time:
            n_result = [is_instance(d, Data) or is_instance(d, dict) for d in data]

        with Timer("Data: check w is_data()") as m_time:
            m_result = [is_data(d) for d in data]

        self.assertEqual(s_result, i_result)
        self.assertEqual(m_result, i_result)
        self.assertEqual(e_result, i_result)
        self.assertEqual(n_result, i_result)

        self.assertGreater(i_time.duration, s_time.duration)
        self.assertGreater(m_time.duration, s_time.duration)

    def test_compare_isinstance_to_text(self):
        num = 1 * 1000 * 1000
        options = {
            0: lambda: 6,
            1: lambda: "string",
            2: lambda: {},
            3: lambda: Data(),
            4: lambda: Null,
        }
        data = [options[randoms.int(len(options))]() for _ in range(num)]
        text_types = (text,)

        with Timer("String: isinstance text_types check") as i_time:
            i_result = [isinstance(d, text_types) for d in data]

        with Timer("String: in text_types check") as s_time:
            s_result = [d.__class__ in text_types for d in data]

        with Timer("String: is check") as e_time:
            e_result = [d.__class__ is text for d in data]

        with Timer("String: is_instance of text check") as n_time:
            n_result = [is_instance(d, text) for d in data]

        with Timer("String: check w is_text method") as m_time:
            m_result = [is_text(d) for d in data]

        # TRY AGAIN
        with Timer("String: isinstance text_types check") as i_time:
            i_result = [isinstance(d, text_types) for d in data]

        with Timer("String: in text_types check") as s_time:
            s_result = [d.__class__ in text_types for d in data]

        with Timer("String: is check") as e_time:
            e_result = [d.__class__ is text for d in data]

        with Timer("String: is_instance of text check") as n_time:
            n_result = [is_instance(d, text) for d in data]

        with Timer("String: check w is_text method") as m_time:
            m_result = [is_text(d) for d in data]

        self.assertEqual(s_result, i_result)
        self.assertEqual(m_result, i_result)
        self.assertEqual(e_result, i_result)
        self.assertEqual(n_result, i_result)

        self.assertGreater(
            i_time.duration.total_seconds() * 1.2,
            s_time.duration.total_seconds(),
            msg="isinstance should be slower than __class__ in set",
        )

        self.assertGreater(
            m_time.duration.total_seconds() * 1.2,
            s_time.duration.total_seconds(),
            "is_text should be slower than isinstance check",
        )

        Log.info(
            "is_text check is {{t|round(places=2)}}x slower than isinstance",
            t=m_time.duration.total_seconds() / i_time.duration.total_seconds(),
        )

    def test_null_compare(self):
        values = randoms.sample([None, Null, {}, Data(), Data(a="b"), [], FlatList()], 1000 * 1000)

        with Timer("is compare") as is_time:
            for v in values:
                v is None

        with Timer("equal compare") as eq_time:
            eq_result = [v == None for v in values]

        with Timer("method compare") as me_time:
            me_result = [is_null(v) for v in values]

        self.assertAlmostEqual(me_result, eq_result, "expecting compare to be the same")

        Log.info(
            "is_null compare is {{t|round(places=2)}}x slower than `== None`",
            t=me_time.duration.seconds / eq_time.duration.seconds,
        )
        Log.info(
            "is_null compare is {{t|round(places=2)}}x slower than `is None`",
            t=me_time.duration.seconds / is_time.duration.seconds,
        )

    def test_from_data(self):
        num = 100 * 1000
        options = {0: Data(), 1: {}, 2: "string", 3: None, 4: Null}
        data = [options[randoms.int(len(options))] for _ in range(num)]

        with Timer("unwrap") as i_time:
            i_result = [from_data(d) for d in data]

    def test_compare_split_replace_vs_lists(self):
        data = []
        for i in range(1_000_000):
            d = "".join(randoms.sample("ab.\b\\", 8))

            try:
                s_result = split_field(d)
                data.append(d)
            except Exception:
                continue

            r_result = split_field_using_replace(d)
            d_result = split_using_double_replace(d)

            self.assertEqual(s_result, r_result, msg=f"problem with {d}")
            self.assertEqual(s_result, d_result, msg=f"problem with {d}")

        with Timer("using standard") as s_time:
            s_result = [split_field(d) for d in data]

        with Timer("using replace") as r_time:
            r_result = [split_field_using_replace(d) for d in data]

        with Timer("using double replace") as d_time:
            d_result = [split_using_double_replace(d) for d in data]

        with Timer("using iterator") as i_time:
            i_result = [split_using_double_replace(d) for d in data]

        # SECOND TIME TO VERIFY SPEED
        with Timer("using standard") as s_time:
            s_result = [split_field(d) for d in data]

        with Timer("using replace") as r_time:
            r_result = [split_field_using_replace(d) for d in data]

        with Timer("using double replace") as d_time:
            d_result = [split_using_double_replace(d) for d in data]

        with Timer("using iterator") as i_time:
            i_result = [split_using_double_replace(d) for d in data]

        self.assertEqual(r_result, s_result)
        self.assertEqual(d_result, s_result)
        self.assertEqual(i_result, s_result)


def is_text(t):
    return t.__class__ is text


def is_instance(d, type_):
    return d.__class__.__name__ == type_.__name__


class Dummy(object):

    __slots__ = ["a"]

    def __init__(self, a):
        self.a = a


def split_using_double_replace(field):
    """
    RETURN field AS ARRAY OF DOT-SEPARATED FIELDS
    """
    try:
        if field.startswith(".."):
            remainder = field.lstrip(".")
            back = len(field) - len(remainder) - 1
            return [".."] * back + [
                k.replace("\a\a", ".").replace("\b", ".") for k in remainder.replace("..", "\a\a").split(".") if k
            ]
        else:
            return [k.replace("\a\a", ".").replace("\b", ".") for k in field.replace("..", "\a\a").split(".") if k]
    except Exception as cause:
        return []


def split_field_using_replace(field):
    """
    RETURN field AS ARRAY OF DOT-SEPARATED FIELDS
    """
    try:
        if field.startswith(".."):
            remainder = field.lstrip(".")
            back = len(field) - len(remainder) - 1
            return [".."] * back + [k.replace("\b", ".") for k in remainder.replace("..", "\b").split(".") if k]
        else:
            return [k.replace("\b", ".") for k in field.replace("..", "\b").split(".") if k]
    except Exception as cause:
        return []


def split_field_using_iterator(field):
    output = deque()
    escape = False
    word = []
    for c in field:
        if escape:
            if c == ".":
                word.append(".")
            else:
                word.append("\\")
                word.append(c)
            escape = False
        else:
            if c == "\\":
                escape = True
            elif c == ".":
                output.append("".join(word))
                word = []
            else:
                word.append(c)
    output.append("".join(word))
    return output



# encoding: utf-8
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this file,
# You can obtain one at http://mozilla.org/MPL/2.0/.
#
# Contact: Kyle Lahnakoski (kyle@lahnakoski.com)
#

from copy import copy, deepcopy

from mo_future import PY3
from mo_testing.fuzzytestcase import FuzzyTestCase

from mo_dots import to_data, Null, listwrap, is_missing, is_null, is_not_null
from mo_dots.lists import last, is_many, FlatList, datawrap

values = [1, 2, 3]


class TestList(FuzzyTestCase):
    def test_generator(self):
        self.assertEqual(last(reversed(values)), 1)

    def test_empty_generator(self):
        self.assertEqual(last(reversed([])), None)

    def test_empty_list(self):
        self.assertEqual(last([]), None)

    def test_empty_set(self):
        self.assertEqual(last(set()), None)

    def test_set(self):
        self.assertEqual(last(set(values)).__class__, int)

    def test_list(self):
        self.assertEqual(last(values), 3)

    def test_flat_list(self):
        v = to_data(values)
        self.assertEqual(last(v), 3)

    def test_string(self):
        v = "test"
        self.assertEqual(last(v), "test")

    def test_bytes(self):
        v = b"test"
        self.assertEqual(last(v), b"test")

    def test_number(self):
        v = 42
        self.assertEqual(last(v), 42)

    def test_object(self):
        v = {}
        self.assertEqual(last(v), to_data({}))

    def test_none(self):
        v = None
        self.assertEqual(last(v), None)

    def test_null(self):
        v = Null
        self.assertEqual(last(v), None)

    def test_is_not_many(self):
        if PY3:
            self.assertEqual(hasattr("", "__iter__"), True)
        self.assertEqual(is_many(""), False)

    def test_index(self):
        v = to_data([{"index": v} for v in values])
        self.assertEqual(v.index, values)

    def test_json(self):
        self.assertEqual(hasattr(to_data(values), "__json__"), False)

    def test_call(self):
        self.assertEqual(hasattr(to_data(values), "__call__"), False)

    def test_listwrap_FlatList(self):
        a = listwrap(values)
        self.assertIsInstance(a, FlatList)
        b = listwrap(a)
        self.assertIs(b, a)

    def test_flatten(self):
        v = listwrap([
            {"a": [{"b": 1}, {"b": 2}, {"b": 3}]},
            {"a": [{"b": 4}, {"b": 5}, {"b": 6}]},
            {"a": [{"b": 7}, {"b": 8}, {"b": 9}]},
        ])

        level1 = v.a
        self.assertEqual(
            level1, [{"b": 1}, {"b": 2}, {"b": 3}, {"b": 4}, {"b": 5}, {"b": 6}, {"b": 7}, {"b": 8}, {"b": 9},],
        )
        self.assertEqual(level1.b, [1, 2, 3, 4, 5, 6, 7, 8, 9])
        self.assertEqual(v.a.b, [1, 2, 3, 4, 5, 6, 7, 8, 9])

        v["a.b"] = 3
        self.assertEqual(v.a.b, [3, 3, 3, 3, 3, 3, 3, 3, 3])

        self.assertEqual(
            v,
            [
                {"a": [{"b": 3}, {"b": 3}, {"b": 3}]},
                {"a": [{"b": 3}, {"b": 3}, {"b": 3}]},
                {"a": [{"b": 3}, {"b": 3}, {"b": 3}]},
            ],
        )

        v.a["b"] = 4
        self.assertEqual(v.a.b, [4, 4, 4, 4, 4, 4, 4, 4, 4])

        v.a.b = 5
        self.assertEqual(v.a.b, [5, 5, 5, 5, 5, 5, 5, 5, 5])

    def test_flatten_w_dot(self):
        v = listwrap([[{"b": 1}, {"b": 2}, {"b": 3}], {"b": 4}, [{"b": 7}, {"b": 8}]])

        level1 = v.get(".")
        self.assertEqual(
            level1, [{"b": 1}, {"b": 2}, {"b": 3}, {"b": 4}, {"b": 7}, {"b": 8}]
        )
        self.assertEqual(level1.b, [1, 2, 3, 4, 7, 8])
        self.assertEqual(v.b, [1, 2, 3, 4, 7, 8])

        v["b"] = 3
        self.assertEqual(v.b, [3, 3, 3, 3, 3, 3])

        self.assertEqual(
            v, [[{"b": 3}, {"b": 3}, {"b": 3}], {"b": 3}, [{"b": 3}, {"b": 3}],],
        )

        v["b"] = 4
        self.assertEqual(v.b, [4, 4, 4, 4, 4, 4])

        v.b = 5
        self.assertEqual(v.b, [5, 5, 5, 5, 5, 5])

    def test_nulls(self):
        self.assertEqual(is_null(None), True)
        self.assertEqual(is_null(Null), True)
        self.assertEqual(is_null(True), False)
        self.assertEqual(is_null(False), False)
        self.assertEqual(is_null(0), False)
        self.assertEqual(is_null(""), False)
        self.assertEqual(is_null({}), False)
        self.assertEqual(is_null([]), False)
        self.assertEqual(is_null(FlatList()), True)
        self.assertEqual(is_null(to_data([0])), False)
        self.assertEqual(is_null(set()), False)
        self.assertEqual(is_null(tuple()), False)

    def test_exists(self):
        self.assertEqual(is_not_null(None), False)
        self.assertEqual(is_not_null(Null), False)
        self.assertEqual(is_not_null(True), True)
        self.assertEqual(is_not_null(False), True)
        self.assertEqual(is_not_null(0), True)
        self.assertEqual(is_not_null(""), True)
        self.assertEqual(is_not_null({}), True)
        self.assertEqual(is_not_null([]), True)
        self.assertEqual(is_not_null(FlatList()), False)
        self.assertEqual(is_not_null(to_data([0])), True)
        self.assertEqual(is_not_null(set()), True)
        self.assertEqual(is_not_null(tuple()), True)

    def test_missing(self):
        self.assertEqual(is_missing(None), True)
        self.assertEqual(is_missing(True), False)
        self.assertEqual(is_missing(False), False)
        self.assertEqual(is_missing(0), False)
        self.assertEqual(is_missing(""), True)
        self.assertEqual(is_missing(Null), True)
        self.assertEqual(is_missing({}), False)
        self.assertEqual(is_missing([]), True)
        self.assertEqual(is_missing(FlatList()), True)
        self.assertEqual(is_missing(to_data([0])), False)
        self.assertEqual(is_missing(set()), True)
        self.assertEqual(is_missing(tuple()), True)

    def test_extend(self):
        v = to_data([])

        v[0] = "a"
        v[5] = "b"
        v[4] = "c"
        v[3] = "d"
        self.assertEqual(v, ["a", None, None, "d", "c", "b"])

    def test_right(self):
        x = to_data([])
        self.assertEqual(x.right(Null), [])
        self.assertEqual(x.right(-3), [])
        self.assertEqual(x.right(-2), [])
        self.assertEqual(x.right(-1), [])
        self.assertEqual(x.right(0), [])
        self.assertEqual(x.right(1), [])
        self.assertEqual(x.right(2), [])
        self.assertEqual(x.right(3), [])

        x = to_data(["a"])
        self.assertEqual(x.right(Null), ["a"])
        self.assertEqual(x.right(-3), [])
        self.assertEqual(x.right(-2), [])
        self.assertEqual(x.right(-1), [])
        self.assertEqual(x.right(0), [])
        self.assertEqual(x.right(1), ["a"])
        self.assertEqual(x.right(2), ["a"])
        self.assertEqual(x.right(3), ["a"])

        x = to_data(["a", "b"])
        self.assertEqual(x.right(Null), ["a", "b"])
        self.assertEqual(x.right(-3), [])
        self.assertEqual(x.right(-2), [])
        self.assertEqual(x.right(-1), [])
        self.assertEqual(x.right(0), [])
        self.assertEqual(x.right(1), ["b"])
        self.assertEqual(x.right(2), ["a", "b"])
        self.assertEqual(x.right(3), ["a", "b"])

    def test_not_right(self):
        x = to_data([])
        self.assertEqual(x.not_right(Null), [])
        self.assertEqual(x.not_right(-3), [])
        self.assertEqual(x.not_right(-2), [])
        self.assertEqual(x.not_right(-1), [])
        self.assertEqual(x.not_right(0), [])
        self.assertEqual(x.not_right(1), [])
        self.assertEqual(x.not_right(2), [])
        self.assertEqual(x.not_right(3), [])

        x = to_data(["a"])
        self.assertEqual(x.not_right(Null), ["a"])
        self.assertEqual(x.not_right(-3), [])
        self.assertEqual(x.not_right(-2), [])
        self.assertEqual(x.not_right(-1), [])
        self.assertEqual(x.not_right(0), ["a"])
        self.assertEqual(x.not_right(1), [])
        self.assertEqual(x.not_right(2), [])
        self.assertEqual(x.not_right(3), [])

        x = to_data(["a", "b"])
        self.assertEqual(x.not_right(Null), ["a", "b"])
        self.assertEqual(x.not_right(-3), [])
        self.assertEqual(x.not_right(-2), [])
        self.assertEqual(x.not_right(-1), [])
        self.assertEqual(x.not_right(0), ["a", "b"])
        self.assertEqual(x.not_right(1), ["a"])
        self.assertEqual(x.not_right(2), [])
        self.assertEqual(x.not_right(3), [])

    def test_left(self):
        x = to_data([])
        self.assertEqual(x.left(Null), [])
        self.assertEqual(x.left(-3), [])
        self.assertEqual(x.left(-2), [])
        self.assertEqual(x.left(-1), [])
        self.assertEqual(x.left(0), [])
        self.assertEqual(x.left(1), [])
        self.assertEqual(x.left(2), [])
        self.assertEqual(x.left(3), [])

        x = to_data(["a"])
        self.assertEqual(x.left(Null), ["a"])
        self.assertEqual(x.left(-3), [])
        self.assertEqual(x.left(-2), [])
        self.assertEqual(x.left(-1), [])
        self.assertEqual(x.left(0), [])
        self.assertEqual(x.left(1), ["a"])
        self.assertEqual(x.left(2), ["a"])
        self.assertEqual(x.left(3), ["a"])

        x = to_data(["a", "b"])
        self.assertEqual(x.left(Null), ["a", "b"])
        self.assertEqual(x.left(-3), [])
        self.assertEqual(x.left(-2), [])
        self.assertEqual(x.left(-1), [])
        self.assertEqual(x.left(0), [])
        self.assertEqual(x.left(1), ["a"])
        self.assertEqual(x.left(2), ["a", "b"])
        self.assertEqual(x.left(3), ["a", "b"])

    def test_not_left(self):
        x = to_data([])
        self.assertEqual(x.not_left(Null), [])
        self.assertEqual(x.not_left(-3), [])
        self.assertEqual(x.not_left(-2), [])
        self.assertEqual(x.not_left(-1), [])
        self.assertEqual(x.not_left(0), [])
        self.assertEqual(x.not_left(1), [])
        self.assertEqual(x.not_left(2), [])
        self.assertEqual(x.not_left(3), [])

        x = to_data(["a"])
        self.assertEqual(x.not_left(Null), ["a"])
        self.assertEqual(x.not_left(-3), [])
        self.assertEqual(x.not_left(-2), [])
        self.assertEqual(x.not_left(-1), [])
        self.assertEqual(x.not_left(0), ["a"])
        self.assertEqual(x.not_left(1), [])
        self.assertEqual(x.not_left(2), [])
        self.assertEqual(x.not_left(3), [])

        x = to_data(["a", "b"])
        self.assertEqual(x.not_left(Null), ["a", "b"])
        self.assertEqual(x.not_left(-3), [])
        self.assertEqual(x.not_left(-2), [])
        self.assertEqual(x.not_left(-1), [])
        self.assertEqual(x.not_left(0), ["a", "b"])
        self.assertEqual(x.not_left(1), ["b"])
        self.assertEqual(x.not_left(2), [])
        self.assertEqual(x.not_left(3), [])

    def lest_last(self):
        self.assertEqual(to_data([]).last(), [])
        self.assertEqual(to_data([]).last(), None)

    def test_filter(self):
        x = to_data(["a", "b"])
        y = x.filter(lambda i: i == "a")
        self.assertEqual(y, ["a"])

    def test_methods(self):
        a = to_data(["a", "b", "c"])

        with self.assertRaises(Exception):
            a.select("b")

        a = to_data(["a", "b", "c"])
        del a[0:1]
        self.assertEqual(a, ["b", "c"])
        a.clear()
        self.assertEqual(a, None)

        a = to_data(["a", "b", "c"])
        del a[1]
        self.assertEqual(a, ["a", "c"])
        a.clear()
        self.assertEqual(a, None)

        a = to_data(["a", "b", "c"])
        self.assertIn("b", a)
        str(a)  # ensure no exception
        self.assertIsNot(list(a), a)

        a = to_data(["a", "b", "c"])
        b = copy(a)
        self.assertIsNot(b, a)
        self.assertEqual(b, a)
        b.remove("c")
        self.assertEqual(b, ["a", "b"])
        self.assertEqual(a, ["a", "b", "c"])

        a = to_data(["a", "b", "c"])
        b = list(a)
        self.assertIsNot(b, a)
        self.assertEqual(b, a)
        b.remove("c")
        self.assertEqual(b, ["a", "b"])
        self.assertEqual(a, ["a", "b", "c"])

        a = to_data(["a", "b", "c"])
        b = a.copy()
        self.assertIsNot(b, a)
        self.assertEqual(b, a)
        b.remove("c")
        self.assertEqual(b, ["a", "b"])
        self.assertEqual(a, ["a", "b", "c"])

        a = to_data(["a", "b", "c"])
        b = deepcopy(a)
        self.assertIsNot(b, a)
        self.assertEqual(b, a)
        b.remove("c")
        self.assertEqual(b, ["a", "b"])
        self.assertEqual(a, ["a", "b", "c"])

    def test_null_hash(self):
        x = to_data([])
        y = Null
        z = None

        a = {x: 42}
        b = {y: 42}
        c = {z: 42}

        self.assertEqual(a, b)
        self.assertEqual(b, c)
        self.assertEqual(c, a)

    def test_eq1(self):
        x = to_data(["a", "b"])
        y = to_data(["a"])

        self.assertFalse(x == y)

    def test_eq2(self):
        x = to_data([Bad(), "b"])
        y = to_data(["a", "b"])

        self.assertFalse(x == y)

    def test_append(self):
        x = to_data(["a", "b"])

        xx = x | x
        self.assertEqual(xx, ["a", "b", "a", "b"])

        zx = "z" + x
        self.assertEqual(zx, ["z", "a", "b"])

        zqx = ("z", "q") + x
        self.assertEqual(zqx, ["z", "q", "a", "b"])

        x += "ok"
        self.assertEqual(x, ["a", "b", "ok"])

    def test_last(self):
        x = to_data([])
        self.assertEqual(x.last(), Null)

    def test_map(self):
        x = to_data(["a", None, "c"])
        self.assertEqual(x.map(lambda c: "::" + c, includeNone=False), ["::a", "::c"])
        self.assertEqual(x.map(lambda c: "::" + c if c is not None else None), ["::a", None, "::c"])

    def test_bad_null(self):
        self.assertFalse(is_null(Bad()))

    def test_to_list(self):
        l = to_data([1, 2, 3])
        result = l.map(lambda x: x * 2).to_list()
        self.assertIsInstance(result, list)
        self.assertEqual(result, [2, 4, 6])

    def test_repr(self):
        d = to_data([1, 2, 3])
        self.assertEqual(repr(d), "to_data([1, 2, 3])")

    def test_int_and_object_gets(self):
        d = to_data([1, 2, 3, {"a": 4, "b": 5}])
        result = d.get("a")
        self.assertEqual(result, [4])


class Bad:
    def __eq__(self, other):
        raise Exception()

    def __req__(self, other):
        raise Exception()


def gen():
    yield "a"
    yield "b"
